@@ -18,20 +18,25 @@ from telebot import types
 from telebot import util
 
 logger = telebot.logger
+
 proxy = None
 
-API_URL = "https://api.telegram.org/bot{0}/{1}"
-FILE_URL = "https://api.telegram.org/file/bot{0}/{1}"
+API_URL = None
+FILE_URL = None
 
 CONNECT_TIMEOUT = 3.5
 READ_TIMEOUT = 9999
 
+CUSTOM_SERIALIZER = None
 
-def _get_req_session():
-    return util.per_thread('req_session', lambda: requests.session())
+ENABLE_MIDDLEWARE = False
 
 
-def _make_request(token, method_name, method='get', params=None, files=None, base_url=API_URL):
+def _get_req_session(reset=False):
+    return util.per_thread('req_session', lambda: requests.session(), reset)
+
+
+def _make_request(token, method_name, method='get', params=None, files=None):
     """
     Makes a request to the Telegram API.
     :param token: The bot's API token. (Created with @BotFather)
@@ -41,7 +46,11 @@ def _make_request(token, method_name, method='get', params=None, files=None, bas
     :param files: Optional files.
     :return: The result parsed to a JSON dictionary.
     """
-    request_url = base_url.format(token, method_name)
+    if API_URL is None:
+        request_url = "https://api.telegram.org/bot{0}/{1}".format(token, method_name)
+    else:
+        request_url = API_URL.format(token, method_name)
+    
     logger.debug("Request: method={0} url={1} params={2} files={3}".format(method, request_url, params, files))
     read_timeout = READ_TIMEOUT
     connect_timeout = CONNECT_TIMEOUT
@@ -99,12 +108,18 @@ def get_file(token, file_id):
 
 
 def get_file_url(token, file_id):
-    method_url = r'getFile'
-    return FILE_URL.format(token, get_file(token, file_id).file_path)
-
+    if FILE_URL is None:
+        return "https://api.telegram.org/file/bot{0}/{1}".format(token, get_file(token, file_id)['file_path'])
+    else:
+        return FILE_URL.format(token, get_file(token, file_id)['file_path'])
+ 
 
 def download_file(token, file_path):
-    url = FILE_URL.format(token, file_path)
+    if FILE_URL is None:
+        url =  "https://api.telegram.org/file/bot{0}/{1}".format(token, file_path)
+    else:
+        url =  FILE_URL.format(token, file_path)
+        
     result = _get_req_session().get(url, proxies=proxy)
     if result.status_code != 200:
         msg = 'The server returned HTTP {0} {1}. Response body:\n[{2}]' \
@@ -114,7 +129,7 @@ def download_file(token, file_path):
 
 
 def send_message(token, chat_id, text, disable_web_page_preview=None, reply_to_message_id=None, reply_markup=None,
-                 parse_mode=None, disable_notification=None):
+                 parse_mode=None, disable_notification=None, timeout=None):
     """
     Use this method to send text messages. On success, the sent Message is returned.
     :param token:
@@ -123,6 +138,8 @@ def send_message(token, chat_id, text, disable_web_page_preview=None, reply_to_m
     :param disable_web_page_preview:
     :param reply_to_message_id:
     :param reply_markup:
+    :param parse_mode:
+    :param disable_notification:
     :return:
     """
     method_url = r'sendMessage'
@@ -137,6 +154,8 @@ def send_message(token, chat_id, text, disable_web_page_preview=None, reply_to_m
         payload['parse_mode'] = parse_mode
     if disable_notification:
         payload['disable_notification'] = disable_notification
+    if timeout:
+        payload['connect-timeout'] = timeout
     return _make_request(token, method_url, params=payload, method='post')
 
 
@@ -240,6 +259,20 @@ def forward_message(token, chat_id, from_chat_id, message_id, disable_notificati
     return _make_request(token, method_url, params=payload)
 
 
+def send_dice(token, chat_id, emoji=None, disable_notification=None, reply_to_message_id=None, reply_markup=None):
+    method_url = r'sendDice'
+    payload = {'chat_id': chat_id}
+    if emoji:
+        payload['emoji'] = emoji
+    if disable_notification:
+        payload['disable_notification'] = disable_notification
+    if reply_to_message_id:
+        payload['reply_to_message_id'] = reply_to_message_id
+    if reply_markup:
+        payload['reply_markup'] = _convert_markup(reply_markup)
+    return _make_request(token, method_url, params=payload)
+
+
 def send_photo(token, chat_id, photo, caption=None, reply_to_message_id=None, reply_markup=None,
                parse_mode=None, disable_notification=None):
     method_url = r'sendPhoto'
@@ -264,7 +297,7 @@ def send_photo(token, chat_id, photo, caption=None, reply_to_message_id=None, re
 
 def send_media_group(token, chat_id, media, disable_notification=None, reply_to_message_id=None):
     method_url = r'sendMediaGroup'
-    media_json, files = _convert_input_media(media)
+    media_json, files = _convert_input_media_array(media)
     payload = {'chat_id': chat_id, 'media': media_json}
     if disable_notification:
         payload['disable_notification'] = disable_notification
@@ -383,6 +416,32 @@ def send_video(token, chat_id, data, duration=None, caption=None, reply_to_messa
     return _make_request(token, method_url, params=payload, files=files, method='post')
 
 
+def send_animation(token, chat_id, data, duration=None, caption=None, reply_to_message_id=None, reply_markup=None,
+               parse_mode=None, disable_notification=None, timeout=None):
+    method_url = r'sendAnimation'
+    payload = {'chat_id': chat_id}
+    files = None
+    if not util.is_string(data):
+        files = {'animation': data}
+    else:
+        payload['animation'] = data
+    if duration:
+        payload['duration'] = duration
+    if caption:
+        payload['caption'] = caption
+    if reply_to_message_id:
+        payload['reply_to_message_id'] = reply_to_message_id
+    if reply_markup:
+        payload['reply_markup'] = _convert_markup(reply_markup)
+    if parse_mode:
+        payload['parse_mode'] = parse_mode
+    if disable_notification:
+        payload['disable_notification'] = disable_notification
+    if timeout:
+        payload['connect-timeout'] = timeout
+    return _make_request(token, method_url, params=payload, files=files, method='post')
+
+
 def send_voice(token, chat_id, voice, caption=None, duration=None, reply_to_message_id=None, reply_markup=None,
                parse_mode=None, disable_notification=None, timeout=None):
     method_url = r'sendVoice'
@@ -512,7 +571,7 @@ def unban_chat_member(token, chat_id, user_id):
 
 def restrict_chat_member(token, chat_id, user_id, until_date=None, can_send_messages=None,
                          can_send_media_messages=None, can_send_other_messages=None,
-                         can_add_web_page_previews=None):
+                         can_add_web_page_previews=None, can_invite_users=None):
     method_url = 'restrictChatMember'
     payload = {'chat_id': chat_id, 'user_id': user_id}
     if until_date:
@@ -525,7 +584,9 @@ def restrict_chat_member(token, chat_id, user_id, until_date=None, can_send_mess
         payload['can_send_other_messages'] = can_send_other_messages
     if can_add_web_page_previews:
         payload['can_add_web_page_previews'] = can_add_web_page_previews
-
+    if can_invite_users:
+        payload['can_invite_users'] = can_invite_users
+        
     return _make_request(token, method_url, params=payload, method='post')
 
 
@@ -618,7 +679,7 @@ def edit_message_text(token, text, chat_id=None, message_id=None, inline_message
         payload['disable_web_page_preview'] = disable_web_page_preview
     if reply_markup:
         payload['reply_markup'] = _convert_markup(reply_markup)
-    return _make_request(token, method_url, params=payload)
+    return _make_request(token, method_url, params=payload, method='post')
 
 
 def edit_message_caption(token, caption, chat_id=None, message_id=None, inline_message_id=None,
@@ -635,7 +696,22 @@ def edit_message_caption(token, caption, chat_id=None, message_id=None, inline_m
         payload['parse_mode'] = parse_mode
     if reply_markup:
         payload['reply_markup'] = _convert_markup(reply_markup)
-    return _make_request(token, method_url, params=payload)
+    return _make_request(token, method_url, params=payload, method='post')
+
+
+def edit_message_media(token, media, chat_id=None, message_id=None, inline_message_id=None, reply_markup=None):
+    method_url = r'editMessageMedia'
+    media_json, file = _convert_input_media(media)
+    payload = {'media': media_json}
+    if chat_id:
+        payload['chat_id'] = chat_id
+    if message_id:
+        payload['message_id'] = message_id
+    if inline_message_id:
+        payload['inline_message_id'] = inline_message_id
+    if reply_markup:
+        payload['reply_markup'] = _convert_markup(reply_markup)
+    return _make_request(token, method_url, params=payload, files=file, method='post' if file else 'get')
 
 
 def edit_message_reply_markup(token, chat_id=None, message_id=None, inline_message_id=None, reply_markup=None):
@@ -649,13 +725,13 @@ def edit_message_reply_markup(token, chat_id=None, message_id=None, inline_messa
         payload['inline_message_id'] = inline_message_id
     if reply_markup:
         payload['reply_markup'] = _convert_markup(reply_markup)
-    return _make_request(token, method_url, params=payload)
+    return _make_request(token, method_url, params=payload, method='post')
 
 
 def delete_message(token, chat_id, message_id):
     method_url = r'deleteMessage'
     payload = {'chat_id': chat_id, 'message_id': message_id}
-    return _make_request(token, method_url, params=payload)
+    return _make_request(token, method_url, params=payload, method='post')
 
 
 # Game
@@ -754,7 +830,8 @@ def send_invoice(token, chat_id, title, description, invoice_payload, provider_t
     :param disable_notification: Sends the message silently. Users will receive a notification with no sound.
     :param reply_to_message_id: If the message is a reply, ID of the original message
     :param reply_markup: A JSON-serialized object for an inline keyboard. If empty, one 'Pay total price' button will be shown. If not empty, the first button must be a Pay button
-    :return: 
+    :param provider_data:
+    :return:
     """
     method_url = r'sendInvoice'
     payload = {'chat_id': chat_id, 'title': title, 'description': description, 'payload': invoice_payload,
@@ -797,7 +874,7 @@ def answer_shipping_query(token, shipping_query_id, ok, shipping_options=None, e
     :param ok: Specify True if delivery to the specified address is possible and False if there are any problems (for example, if delivery to the specified address is not possible)
     :param shipping_options: Required if ok is True. A JSON-serialized array of available shipping options.
     :param error_message: Required if ok is False. Error message in human readable form that explains why it is impossible to complete the order (e.g. "Sorry, delivery to your desired address is unavailable'). Telegram will display this message to the user.
-    :return: 
+    :return:
     """
     method_url = 'answerShippingQuery'
     payload = {'shipping_query_id': shipping_query_id, 'ok': ok}
@@ -815,7 +892,7 @@ def answer_pre_checkout_query(token, pre_checkout_query_id, ok, error_message=No
     :param pre_checkout_query_id: Unique identifier for the query to be answered
     :param ok: Specify True if everything is alright (goods are available, etc.) and the bot is ready to proceed with the order. Use False if there are any problems.
     :param error_message: Required if ok is False. Error message in human readable form that explains the reason for failure to proceed with the checkout (e.g. "Sorry, somebody just bought the last of our amazing black T-shirts while you were busy filling out your payment details. Please choose a different color or garment!"). Telegram will display this message to the user.
-    :return: 
+    :return:
     """
     method_url = 'answerPreCheckoutQuery'
     payload = {'pre_checkout_query_id': pre_checkout_query_id, 'ok': ok}
@@ -921,6 +998,49 @@ def delete_sticker_from_set(token, sticker):
     return _make_request(token, method_url, params=payload, method='post')
 
 
+def send_poll(token, chat_id, poll, explanation_parse_mode=None, disable_notifications=False, reply_to_message_id=None, reply_markup=None):
+    method_url = r'sendPoll'
+    payload = {
+        'chat_id': str(chat_id),
+        'question': poll.question,
+        'options': _convert_list_json_serializable(poll.options)}
+
+    if poll.is_anonymous is not None:
+        payload['is_anonymous'] = poll.is_anonymous
+    if poll.type is not None:
+        payload['type'] = poll.type
+    if poll.allows_multiple_answers is not None:
+        payload['allows_multiple_answers'] = poll.allows_multiple_answers
+    if poll.correct_option_id is not None:
+        payload['correct_option_id'] = poll.correct_option_id
+    if poll.explanation is not None:
+        payload['explanation'] = poll.explanation
+    if explanation_parse_mode is not None:
+        payload['explanation_parse_mode'] = explanation_parse_mode
+    if poll.open_period is not None:
+        payload['open_period'] = poll.open_period
+    if poll.close_date is not None:
+        payload['close_date'] = poll.close_date
+    if poll.is_closed is not None:
+        payload['is_closed'] = poll.is_closed
+
+    if disable_notifications:
+        payload['disable_notification'] = disable_notifications
+    if reply_to_message_id is not None:
+        payload['reply_to_message_id'] = reply_to_message_id
+    if reply_markup is not None:
+        payload['reply_markup'] = _convert_markup(reply_markup)
+    return _make_request(token, method_url, params=payload)
+
+
+def stop_poll(token, chat_id, message_id, reply_markup=None):
+    method_url = r'stopPoll'
+    payload = {'chat_id': str(chat_id), 'message_id': message_id}
+    if reply_markup:
+        payload['reply_markup'] = _convert_markup(reply_markup)
+    return _make_request(token, method_url, params=payload)
+
+
 def _convert_list_json_serializable(results):
     ret = ''
     for r in results:
@@ -937,11 +1057,17 @@ def _convert_markup(markup):
     return markup
 
 
-def _convert_input_media(array):
+def _convert_input_media(media):
+    if isinstance(media, types.InputMedia):
+        return media._convert_input_media()
+    return None, None
+
+
+def _convert_input_media_array(array):
     media = []
     files = {}
     for input_media in array:
-        if isinstance(input_media, types.JsonSerializable):
+        if isinstance(input_media, types.InputMedia):
             media_dict = input_media.to_dic()
             if media_dict['media'].startswith('attach://'):
                 key = media_dict['media'].replace('attach://', '')
